import os
import glob
import logging
import shutil
import sqlite3
import time


class FusionHistoryStorage():
    # Fusion means both file (stdout and stderr) and sqlite storage
    STDOUT_LINES_LIMIT = 50

    CREATE_TABLES_SQL = ['''
        CREATE TABLE IF NOT EXISTS TASK_RUNS
            (   task_run_id INTEGER PRIMARY KEY AUTOINCREMENT,
                started_at TEXT NOT NULL,
                finished_at TEXT,
                task_id TEXT,
                pid INTEGER,
                exit_code INTEGER );''',
                         "CREATE INDEX IF NOT EXISTS task_id ON task_runs (task_id);",
                         "CREATE INDEX IF NOT EXISTS started_at ON task_runs (started_at);",
                         "CREATE INDEX IF NOT EXISTS task_id_started_at ON task_runs (task_id, started_at);",
                         "CREATE INDEX IF NOT EXISTS finished_at ON task_runs (finished_at);",
                         "CREATE INDEX IF NOT EXISTS started_at_exit_code ON task_runs (started_at, exit_code);"
                         ]

    COLUMNS = ['task_run_id', 'started_at', 'finished_at', 'task_id',
               'pid', 'exit_code']

    def __init__(self, storage_dir="./var"):
        self.storage_dir = storage_dir
        if not os.path.isdir(self.storage_dir):
            os.makedirs(self.storage_dir)
        self.sqlite_connection = sqlite3.connect(self.storage_dir + '/ereb.db')
        logging.info('FusionHistoryStorage => Database connected')
        for sql in self.CREATE_TABLES_SQL:
            self.sqlite_connection.execute(sql)
        logging.info('FusionHistoryStorage => Tables created')

    def get_recent_failed_task_runs(self, limit):
        columns = ['task_id', 'task_run_id', 'started_at', 'duration']
        result = self.select_to_dict('''
            select task_id, task_run_id, started_at,
                strftime('%s', finished_at) - strftime('%s', started_at) as duration
            from task_runs
            where exit_code != 0
            and finished_at != 'None'
            and started_at > datetime('now', '-1 month')
            order by started_at desc
            limit {}
        '''.format(limit), columns)

        return result

    def get_currently_running_tasks(self):
        result = self.select_to_dict('''
            select *
            from task_runs
            where finished_at = 'None'
        ''')

        return result

    def get_last_day_for_task_id(self, task_id):
        result = self.select_to_dict('''
            select *
            from task_runs
            where task_id = '%s'
            order by task_run_id desc
            limit 1
        ''' % task_id)

        if len(result) > 0:
            last_day = result[0]['started_at'][:10]
        else:
            last_day = None

        return last_day

    def get_task_runs_for_task_id(self, task_id, limit=20):
        result = self.select_to_dict('''
            select *
            from task_runs
            where task_id = '%s'
            order by task_run_id desc
            limit %s
        ''' % (task_id, limit))
        return result

    def get_task_stats_for_dashboard(self):
        columns = ['task_id', 'duration_avg', 'fails']
        start_time = time.time()
        failed_tasks = self.select_to_dict('''
        select
            task_id,
            round(avg(strftime('%s', finished_at) - strftime('%s', started_at)), 2) as duration_avg,
            sum(case when exit_code != 0 then 1 else 0 end) as fails

        from task_runs
            where started_at > datetime('now', '-1 day')
            group by task_id
            order by duration_avg desc;
        ''', columns)

        logging.warning('FusionHistoryStorage :: get_task_stats_for_dashboard sql time %s', time.time() - start_time)
        # Sqlite3 is single threaded in it takes fucking LONG time to execute 4 queries one by one
        # that's why
        # this query is ugly, single and works 4x time faster than previous

        result = {}
        result['slow_tasks'] = failed_tasks[:5]
        result['failed_tasks'] = sorted(failed_tasks, key=lambda x: x['fails'], reverse=True)[:5]

        return result

    def get_task_list_stats(self):
        columns = ['task_id', 'duration_avg', 'duration_min', 'duration_max',
                   'success', 'error', 'exit_codes']
        task_stats = self.select_to_dict('''
            select  task_id,
            round(avg(strftime('%s', finished_at) - strftime('%s', started_at)), 2) as duration_avg,
            min(strftime('%s', finished_at) - strftime('%s', started_at)) as duration_min,
            max(strftime('%s', finished_at) - strftime('%s', started_at)) as duration_max,
            sum(case when exit_code = 0 then 1 else 0 end) as success,
            sum(case when exit_code != 0 then 1 else 0 end) as error,
            group_concat(exit_code) as exit_codes
            from task_runs
            where started_at > datetime('now', '-1 day')
            group by task_id
        ''', columns)

        stats_by_task_id = {}
        for task in task_stats:
            if task['exit_codes'] != '':
                task['exit_codes'] = str(task['exit_codes']).split(',')[-self.STDOUT_LINES_LIMIT:]
            else:
                task.pop('exit_codes')

            stats_by_task_id[task['task_id']] = task

        return stats_by_task_id

    def get_detailed_task_run_info(self, task_id, task_run_id, content_size_limit=True):
        task_run = {}
        task_run['state'] = self.select_to_dict('''
            select *
            from task_runs
            where task_run_id = '%s'
        ''' % task_run_id)[0]

        day = task_run['state']['started_at'][:10]
        task_run_path = self.storage_dir + '/%s/%s/%s' % (task_id, day, task_run_id)

        for x in ['stdout', 'stderr']:
            file_path = '/'.join([task_run_path, x])
            if os.path.isfile(file_path):
                with open(task_run_path + '/' + x) as file_content:
                    lines = file_content.read().split("\n")
                    if content_size_limit:
                        task_run[x] = "\n".join(lines[-50:])
                    else:
                        task_run[x] = "\n".join(lines)
            else:
                task_run[x] = ''

        return task_run

    def finalize_task_run(self, task_run):
        task_run.finalize()
        task_run.state['exit_code'] = '-1'
        self.update_state_for_task_run(task_run)

    def get_task_run_path(self, task_run):
        return '/'.join([
            self.storage_dir,
            task_run.task_id,
            task_run.state['started_at'][:10],
            task_run.id
        ])

    def update_state_for_task_run(self, task_run):
        self.sqlite_connection.execute('''
            update task_runs
            set started_at = '%s',
                finished_at = '%s',
                task_id = '%s',
                pid = '%s',
                exit_code = '%s'
            where task_run_id = '%s';''' % (
            task_run.state['started_at'], task_run.state['finished_at'],
            task_run.state['task_id'], task_run.state['pid'],
            task_run.state['exit_code'], task_run.id))
        self.sqlite_connection.commit()
        logging.warn('FusionHistoryStorage updated!')

    def update_stdout_for_task_run_id(self, task_run):
        stdout_file_path = self.get_task_run_path(task_run) + '/stdout'
        self.write_to_file(stdout_file_path, task_run.stdout)

    def update_stderr_for_task_run_id(self, task_run):
        stderr_file_path = self.get_task_run_path(task_run) + '/stderr'
        self.write_to_file(stderr_file_path, task_run.stderr)

    def task_valid_to_run(self, task_id):
        currently_running = self.select_to_dict('''
            select *
            from task_runs
            where task_id = '%s' and finished_at = 'None'
        ''' % task_id)
        return len(currently_running) == 0

    def prepare_task_run(self, task_run):
        self.sqlite_connection.execute('''
            insert into task_runs(task_id, started_at)
            values ('%s', '%s');
        ''' % (task_run.task_id, task_run.state['started_at']))
        self.sqlite_connection.commit()

        task_run_id = str(list(self.sqlite_connection.execute('select last_insert_rowid();'))[0][0])

        task_run.id = task_run_id

        task_path = '/'.join([self.storage_dir, task_run.task_id])
        task_path = self.get_task_run_path(task_run)
        if not os.path.isdir(task_path):
            os.makedirs(task_path)
        self.remove_old_history_for_task_id(task_run.task_id, days_limit=30)

    def remove_old_history_for_task_id(self, task_id, days_limit=30):
        day_dirs = glob.glob(self.storage_dir + '/' + task_id + '/*')
        if len(day_dirs) > days_limit:
            day_dirs.sort()
            last_day = day_dirs[0]
            shutil.rmtree(last_day)
        self.sqlite_connection.execute('''
            delete from task_runs
            where task_id = '%s' and started_at < datetime('now', '-%d days');
        ''' % (task_id, days_limit))
        self.sqlite_connection.commit()

    def last_fails(self, task_id):
        task_runs = self.select_to_dict('''
            select *
            from task_runs
            where task_id = '%s'
            order by started_at desc
        ''' % task_id)
        result = 0
        for task_run in task_runs:
            if task_run['exit_code'] == 0:
                return result
            else:
                result += 1
        return result

    def select_to_dict(self, sql, columns=COLUMNS):
        cursor = self.sqlite_connection.execute(sql)
        result = []

        for row in cursor:
            r = {}
            for i, column in enumerate(columns):
                r[column] = row[i]
            result.append(r)

        return result

    def write_to_file(self, path, content):
        with open(path, 'w') as f:
            f.write(content)



from tornado.ioloop import IOLoop
import tornado.web
import json
import logging
from tornado import websocket
import subprocess
from functools import partial
import sys
import signal
import pkg_resources
import os
from tornado.options import define, options

from ereb.tasks_controller import TaskController


class SocketHandler(websocket.WebSocketHandler):

    def initialize(self, task_controller, websocket_clients):
        self.task_controller = task_controller
        self.websocket_clients = websocket_clients

    def check_origin(self, origin):
        return True

    def open(self):
        self.websocket_clients.append(self)
        logging.info('WebSocket opened, clients: %s' % len(self.websocket_clients))
        self.write_message(self.task_controller.get_status())

    def on_close(self):
        self.websocket_clients.remove(self)
        logging.info('WebSocket closed, clients: %s' % len(self.websocket_clients))


class GenericTasksHandler(tornado.web.RequestHandler):

    def initialize(self, task_controller):
        self.task_controller = task_controller

    def post(self):
        params = json.loads(self.request.body.decode())
        self.task_controller.run_generic_task(params['name'], params['cmd'], params['timeout'])
        self.write('ok')


class TasksHandler(tornado.web.RequestHandler):

    def initialize(self, task_controller):
        self.task_controller = task_controller

    def get(self, task_id, action, task_run_id):
        result = '404'
        if task_id == '' and action == '':
            tasks = self.task_controller.get_task_list(with_history=True, task_run_limit=20)
            result = json.dumps(tasks)
        elif task_id != '':
            task = self.task_controller.get_task_by_id(task_id, with_extra_info=True)
            if not task:
                self.raise_404('task %s not found' % task_id)
            if action == '':
                result = json.dumps({
                    'config': task,
                    'runs': self.task_controller.get_task_runs_for_task_id(task_id)
                })
            elif action == 'run':
                self.task_controller.run_task_by_task_id(task_id)
                result = 'task_run'
            elif action == 'shutdown':
                self.task_controller.shutdown_run_for_task_id(task_id)
                result = 'task %s killed' % task_id
            elif action == 'task_runs' and task_run_id != '':
                detailed_task_run = self.task_controller.get_detailed_task_run_info(task_id, task_run_id)
                if not detailed_task_run:
                    self.raise_404('Task run %s/%s not found' % (task_id, task_run_id))
                else:
                    result = json.dumps(detailed_task_run)
            elif action == 'task_run_stdout' and task_run_id != '':
                task_run_stdout = self.task_controller.get_stdout_for_task_run_id(task_id, task_run_id)
                if not task_run_stdout:
                    self.raise_404('Stdout %s/%s not found' % (task_id, task_run_id))
                else:
                    result = task_run_stdout
            elif action == 'task_run_stderr' and task_run_id != '':
                task_run_stderr = self.task_controller.get_stderr_for_task_run_id(task_id, task_run_id)
                if not task_run_stderr:
                    self.raise_404('Stderr for %s/%s not found' % (task_id, task_run_id))
                else:
                    result = task_run_stderr

        self.set_header('Access-Control-Allow-Origin', '*')
        self.write(result)

    def post(self, task_id, action, task_run_id):
        result = '404'
        config = json.loads(self.request.body.decode())

        if task_id == '' and action == '':
            is_valid_config = self.task_controller.validate_config(config)
            if not is_valid_config:
                self.raise_500('Cannot update task with config: %s' % json.dumps(config))
            else:
                task_id = config['task_id'].replace(' ', '_')
                task = self.task_controller.set_task_by_id(task_id, config)
                result = 'Success'
                self.set_header('Access-Control-Allow-Origin', '*')
                self.write(result)
        elif task_id != '':
            is_valid_config = self.task_controller.validate_config(config)

            if not is_valid_config:
                self.raise_500('Cannot update task with config: %s' % json.dumps(config))
            else:
                task = self.task_controller.set_task_by_id(task_id, config)
                result = 'Success'
                self.set_header('Access-Control-Allow-Origin', '*')
                self.write(result)

    def delete(self, task_id, action, task_run_id):
        result = '404'

        if task_id == '' and action == '':
            self.raise_404('Unknown route')
        elif task_id != '':
            task = self.task_controller.delete_task_by_id(task_id)
            if not task:
                self.raise_500('Cannot delete task with config: %s' % json.dumps(config))
            else:
                result = 'Success'
        self.set_header('Access-Control-Allow-Origin', '*')
        self.write(result)

    def options(self, task_id, action, task_run_id):
        self.set_header('Access-Control-Allow-Origin', '*')
        self.set_header('Access-Control-Allow-Methods', 'POST, GET, OPTIONS, DELETE')

        self.write('')

    def raise_404(self, message):
        self.clear()
        self.set_status(404)
        self.set_header('Access-Control-Allow-Origin', '*')
        self.finish(message)

    def raise_500(self, message):
        self.clear()
        self.set_status(500)
        self.set_header('Access-Control-Allow-Origin', '*')
        self.finish(message)


class RunnerHandler(tornado.web.RequestHandler):

    def initialize(self, task_controller):
        self.task_controller = task_controller

    def get(self, cmd):
        if cmd == '':
            result = json.dumps(self.task_controller.get_status())
            self.set_header('Access-Control-Allow-Origin', '*')
            self.write(result)
        if cmd == 'dashboard':
            result = json.dumps(self.task_controller.get_dashboard_info())
            self.set_header('Access-Control-Allow-Origin', '*')
            self.write(result)
        if cmd == 'recent_failed_tasks':
            recent_tasks = self.task_controller.get_recent_history(100)
            failed_tasks = []
            for task in recent_tasks:
                if task['exit_code'] != 0 and task['exit_code'] != 'None':
                    failed_tasks.append(task)
            result = json.dumps(failed_tasks)
            self.set_header('Access-Control-Allow-Origin', '*')
            self.write(result)
        elif cmd == 'start':
            self.task_controller.start_task_loop()
            self.set_header('Access-Control-Allow-Origin', '*')
            self.write('Success')
        elif cmd == 'stop':
            self.task_controller.stop_task_loop()
            self.set_header('Access-Control-Allow-Origin', '*')
            self.write('Success')


def shutdown(shutdown_tasks, *args):
    logging.info("shutting down...")
    shutdown_tasks()
    IOLoop.current().stop()
    sys.exit(0)


def main():
    define("port", default=8888, type=int, help="port to listen on")
    define("tasks_dir", default=os.path.dirname(os.path.realpath(__file__)) +
           "/../etc", type=str, help="directory with tasks config files")
    define("history_dir", default=os.path.dirname(os.path.realpath(__file__)) +
           "/../var", type=str, help="directory for history storage")
    define("notifier_config", default=os.path.dirname(os.path.realpath(__file__)) +
           "/../notifier.json", type=str, help="notifier json config")
    define("notify_to", default="logger", type=str, help="notifications channel")
    define("notifier_host", default="hostname", type=str, help="host for links in notifications")
    define("datadog_metrics", default=False, help="send metrics to datadog")
    define("datadog_agent_host", default="localhost", help="host where datadog agent is running")
    define("datadog_agent_port", default=8125, help="port where datadog agent is running")
    define("datadog_tag", default="default", help="value for special `ereb:%datadog_tag%` tag in datadog metrics, sent with every event")

    tornado.options.parse_command_line()

    datadog_config = {
        'enabled': False
    }
    if options.datadog_metrics:
        from datadog import initialize, statsd
        initialize(statsd_host=options.datadog_agent_host, statsd_port=options.datadog_agent_port)
        datadog_config = {
            'enabled': True,
            'statsd_client': statsd,
            'tag': "ereb:%s" % options.datadog_tag
        }

    datadog_metrics = True if options.datadog_metrics else False


    ereb_version = version = pkg_resources.require("ereb")[0].version

    default_wi_config = """
        window.DEFAULT_CONFIG = {}
    """.format({
        'port': options.port,
        'version': ereb_version
    })

    default_wi_config_path = os.path.dirname(os.path.realpath(__file__)) + '/ereb-wi/default_config.js'

    with open(default_wi_config_path, 'w') as f:
        f.write(default_wi_config)

    try:
        with open(options.notifier_config) as notifier_config_file:
            notifier_config = json.load(notifier_config_file)
    except Exception as e:
        logging.info('You can use Slack notifier by create notifier.json')
        notifier_config = {}

    websocket_clients = []

    task_controller = TaskController(
        tasks_dir=options.tasks_dir,
        history_dir=options.history_dir,
        notifier_config=notifier_config,
        notify_to=options.notify_to,
        notifier_host=options.notifier_host,
        port=options.port,
        websocket_clients=websocket_clients,
        datadog_config=datadog_config)

    logging.info("Starting EREB on http://{}:{}".format('0.0.0.0', options.port))
    if datadog_metrics:
        logging.info("Sending metrics to Datadog enabled")

    task_controller.start()

    application = tornado.web.Application([
        (r"/tasks/?([^/]*)/?([^/]*)/?([^/]*)$", TasksHandler, dict(task_controller=task_controller)),
        (r"/generic_tasks/run", GenericTasksHandler, dict(task_controller=task_controller)),
        (r"/status/?(.*)$", RunnerHandler, dict(task_controller=task_controller)),
        (r'/ws', SocketHandler, dict(task_controller=task_controller, websocket_clients=websocket_clients)),
        (r"/(.*)", tornado.web.StaticFileHandler,
         {"path": os.path.dirname(os.path.realpath(__file__)) + "/ereb-wi", "default_filename": "index.html"})
    ], gzip=True)

    signal.signal(signal.SIGTERM, partial(shutdown, task_controller.shutdown_tasks))
    signal.signal(signal.SIGHUP, partial(shutdown, task_controller.shutdown_tasks))
    signal.signal(signal.SIGINT, partial(shutdown, task_controller.shutdown_tasks))

    application.listen(options.port)
    IOLoop.instance().start()


if __name__ == "__main__":
    main()



import os
import json
import psutil
from tornado.ioloop import PeriodicCallback
import logging
import re
import datetime
from os.path import isfile

from ereb.tasks_scheduler import TasksScheduler
from ereb.fusion_history_storage import FusionHistoryStorage
from ereb.task_run import TaskRun
from ereb.notifier import Notifier


class TaskController():

    def __init__(self, tasks_dir="etc", history_dir="./var",
                 notifier_config={}, notify_to='logger', notifier_host='hostname',
                 websocket_clients=[], port=8888, datadog_config={}):
        self.tasks_dir = tasks_dir
        if not os.path.exists(self.tasks_dir):
            os.makedirs(self.tasks_dir)

        self.websocket_clients = websocket_clients
        self.history_storage = FusionHistoryStorage(history_dir)
        self.notifier = Notifier(notifier_config=notifier_config,
                                 notify_to=notify_to,
                                 websocket_clients=websocket_clients,
                                 host=notifier_host,
                                 port=port)
        self.task_scheduler = TasksScheduler(tasks_dir=tasks_dir,
                                             history_storage=self.history_storage,
                                             notifier=self.notifier, datadog_config=datadog_config)
        self.check_processes()
        self.process_checking_loop = PeriodicCallback(self.check_processes, 10000)
        self.process_checking_loop.start()

    def shutdown_tasks(self, tasks=None):
        tasks = tasks or self.history_storage.get_currently_running_tasks()
        for task_state in tasks:
            task_run = TaskRun.from_state(task_state)
            task_run.shutdown()

    def shutdown_run_for_task_id(self, task_id):
        tasks = self.history_storage.get_currently_running_tasks()
        for task_state in tasks:
            if task_state['task_id'] == task_id:
                TaskRun.from_state(task_state).shutdown()

    def update_config(self):
        return self.task_scheduler.update_config()

    def start(self):
        self.task_scheduler.start()

    def start_task_loop(self):
        self.task_scheduler.start_task_loop()

    def stop_task_loop(self):
        self.task_scheduler.stop_task_loop()

    def get_status(self):
        return self.task_scheduler.get_status()

    def check_processes(self):
        logging.info('Checking processes')
        for currently_running_task in self.history_storage.get_currently_running_tasks():
            task_run = TaskRun.from_state(currently_running_task)

            if 'pid' in task_run.state:
                if task_run.state['pid'] != 'None':
                    if psutil.pid_exists(task_run.state['pid']):
                        logging.info('Task %s with run %s is alive', task_run.task_id, task_run.id)
                    else:
                        logging.info('Task %s with run %s is dead already; finalized', task_run.task_id, task_run.id)
                        self.history_storage.finalize_task_run(task_run)
                        # FIXME: Prolly it's now working. DO SOMETHING
                else:
                    logging.info('Task %s with run %s is in unknown state, no pid; finalized',
                                 task_run.task_id, task_run.id)
                    self.history_storage.finalize_task_run(task_run)
                    # FIXME: Prolly it's now working. DO SOMETHING

    def get_next_tasks(self):
        return self.task_scheduler.get_next_tasks()

    def get_task_list(self, with_history=False, task_run_limit=20):
        task_list = self.task_scheduler.tasks_list

        task_stats = self.history_storage.get_task_list_stats()

        for task in task_list:
            if with_history:
                if task['name'] in task_stats:
                    task['stats'] = task_stats.get(task['name'])

        return sorted(task_list, key=lambda x: x['name'])

    def get_dashboard_info(self):
        dashboard = {}
        dashboard['recent_fails'] = self.history_storage.get_recent_failed_task_runs(20)
        dashboard.update(self.history_storage.get_task_stats_for_dashboard())

        return dashboard

    def get_task_by_id(self, task_id, with_extra_info=False):
        return self.task_scheduler.get_task_by_id(task_id, with_extra_info)

    def set_task_by_id(self, task_id, task_config):
        # try to update task first
        task = self.get_task_by_id(task_id)
        if task:
            task.update(task_config)
            task_config = task

        f = open(self.tasks_dir + '/%s.json' % task_id, 'w')
        f.write(json.dumps(task_config, sort_keys=True, indent=4))
        f.close()
        self.task_scheduler.reschedule_tasks()
        return True

    def delete_task_by_id(self, task_id):
        f = self.tasks_dir + '/%s.json' % task_id
        os.remove(f)
        self.task_scheduler.check_config()
        return True

    def get_task_runs_for_task_id(self, task_id, limit=20):
        return self.history_storage.get_task_runs_for_task_id(task_id, limit)

    def get_detailed_history_for_task_id(self, task_id, limit=20):
        return self.history_storage.get_detailed_history_for_task_id(task_id, limit)

    def get_detailed_task_run_info(self, task_id, task_run_id):
        return self.history_storage.get_detailed_task_run_info(task_id, task_run_id)

    def get_stdout_for_task_run_id(self, task_id, task_run_id):
        detailed_task = self.history_storage.get_detailed_task_run_info(task_id, task_run_id, content_size_limit=False)
        return detailed_task['stdout']

    def get_stderr_for_task_run_id(self, task_id, task_run_id):
        detailed_task = self.history_storage.get_detailed_task_run_info(task_id, task_run_id, content_size_limit=False)
        return detailed_task['stderr']

    def run_task_by_task_id(self, task_id):
        task = self.get_task_by_id(task_id)
        logging.info('MANUAL RUN | Running %s task' % task['name'])
        self.task_scheduler.run_task_by_name_and_cmd(
            task['name'], task['cmd'], task.get('timeout', -1), task.get('fails_before_notify', 0))

    def run_generic_task(self, name, cmd, timeout):
        logging.info('GENERIC RUN | %s | %s | %s' % (name, cmd, timeout))
        _name = "__generic_" + name
        self.task_scheduler.run_task_by_name_and_cmd(_name, cmd, timeout)

    def get_tasks_config(self):
        return self.task_scheduler.get_tasks_config()

    def validate_config(self, config):
        return self.task_scheduler.validate_config(config)



import datetime
import signal
import psutil
import logging


def kill_pid(pid, sig=signal.SIGTERM):
    try:
        parent = psutil.Process(pid)
    except psutil.NoSuchProcess:
        return
    children = parent.children(recursive=True)
    for process in children:
        process.send_signal(sig)
        try:
            process.send_signal(sig)
        except ProcessLookupError:
            continue
        except psutil.NoSuchProcess:
            continue

        try:
            logging.info("couldn't kill process by TERM signal, let's start use KILL")
            process.send_signal(signal.SIGKILL)
        except psutil.NoSuchProcess:
            return
    parent.send_signal(sig)


class TaskRun():

    def __init__(self, task_id):
        self.started_at = datetime.datetime.utcnow()
        self.finished_at = None
        self.task_id = task_id
        self.id = None

        self.state = {
            'current': 'running',
            'started_at': self.started_at.strftime('%Y-%m-%d %H:%M:%S'),
            'finished_at': None,
            'pid': None,
            'exit_code': None,
            'task_id': self.task_id
        }
        self.stdout = ''
        self.stderr = ''

    @classmethod
    def from_state(klass, state):
        task_run = klass(state['task_id'])
        task_run.id = state['task_run_id']
        task_run.state = state

        return task_run

    def shutdown(self):
        kill_pid(self.state['pid'])

    def started_at(self):
        return datetime.datetime.strptime(self.state['started_at'], '%Y-%m-%d %H:%M:%S')

    def finalize(self):
        self.finished_at = datetime.datetime.utcnow()
        self.state['finished_at'] = self.finished_at.strftime('%Y-%m-%d %H:%M:%S')
        self.state['current'] = 'finished'

    def log_info(self, fails, lines_count=2):
        def get_trimmed_text(txt, lines_count=2):
            splitted = list(filter(None, txt.split('\n')))
            data = {'first': '\n'.join(splitted[:lines_count])}
            del splitted[:lines_count]
            data['last'] = '\n'.join(splitted[-lines_count:])
            return data

        stdout = get_trimmed_text(self.stdout, lines_count)
        stderr = get_trimmed_text(self.stderr, lines_count)
        return {
            'stdout_first': stdout['first'],
            'stdout_last': stdout['last'],
            'stderr_first': stderr['first'],
            'stderr_last': stderr['last'],
            'run_time': str(datetime.datetime.utcnow() - self.started_at),
            'exit_code': self.state['exit_code'],
            'task_id': self.task_id,
            'fails': str(fails)
        }



from tornado.process import Subprocess
from tornado.iostream import PipeIOStream


class AASubprocess(Subprocess):
    """
    Extension to tornado.process.Subprocess to support
    - process timeout
    - pass stdin bytes
    - real-time stdout data chunk callback
    - real-time stderr data chunk callback
    - process end callback
    """
    def __init__(
                self,
                command,
                timeout=-1,
                stdout_chunk_callback=None,
                stderr_chunk_callback=None,
                exit_process_callback=None,
                stdin_bytes=None,
                io_loop=None,
                kill_on_timeout=False
                ):
        """
        Initializes the subprocess with callbacks and timeout.

        :param command: command like ['java', '-jar', 'test.jar']
        :param timeout: timeout for subprocess to complete, if negative or zero then no timeout
        :param stdout_chunk_callback: callback(bytes_data_chuck_from_stdout)
        :param stderr_chunk_callback: callback(bytes_data_chuck_from_stderr)
        :param exit_process_callback: callback(exit_code, was_expired_by_timeout)
        :param stdin_bytes: bytes data to send to stdin
        :param io_loop: tornado io loop on None for current
        :param kill_on_timeout: kill(-9) or terminate(-15)?
        """
        self.aa_exit_process_callback = exit_process_callback
        self.aa_kill_on_timeout = kill_on_timeout
        stdin = Subprocess.STREAM if stdin_bytes else None
        stdout = Subprocess.STREAM if stdout_chunk_callback else None
        stderr = Subprocess.STREAM if stderr_chunk_callback else None

        Subprocess.__init__(self, command, stdin=stdin, stdout=stdout, stderr=stderr, io_loop=io_loop, shell=True)

        self.aa_process_expired = False
        self.aa_terminate_timeout = self.io_loop.call_later(timeout, self.aa_timeout_callback) if timeout > 0 else None

        self.set_exit_callback(self.aa_exit_callback)

        if stdin:
            self.stdin.write(stdin_bytes)
            self.stdin.close()

        if stdout:
            output_stream = PipeIOStream(self.stdout.fileno())

            def on_stdout_chunk(data):
                stdout_chunk_callback(data)
                if not output_stream.closed():
                    output_stream.read_bytes(102400, on_stdout_chunk, None, True)

            output_stream.read_bytes(102400, on_stdout_chunk, None, True)

        if stderr:
            stderr_stream = PipeIOStream(self.stderr.fileno())

            def on_stderr_chunk(data):
                stdout_chunk_callback(data)
                if not stderr_stream.closed():
                    stderr_stream.read_bytes(102400, on_stderr_chunk, None, True)

            stderr_stream.read_bytes(102400, on_stderr_chunk, None, True)

    def aa_timeout_callback(self):
        if self.aa_kill_on_timeout:
            self.proc.kill()
        else:
            self.proc.terminate()
        self.aa_process_expired = True

    def aa_exit_callback(self, status):
        if self.aa_terminate_timeout:
            self.io_loop.remove_timeout(self.aa_terminate_timeout)
        # need to post this call to make sure it is processed AFTER all outputs
        if self.aa_exit_process_callback:
            self.io_loop.add_callback(self.aa_exit_process_callback, status, self.aa_process_expired)



import logging

from ereb.task_run import TaskRun
from ereb.aa_subprocess import AASubprocess
from datadog import statsd


class TaskRunner():

    def __init__(self, task_id, history_storage, notifier, on_error_callback, datadog_config):
        self.task_id = task_id
        self.history_storage = history_storage
        self.notifier = notifier
        self.on_error_callback = on_error_callback
        self.datadog_config = datadog_config

    def run_task(self, cmd, timeout=-1, fails_before_notify=0):
        logging.info("Runner started, %s with timeout %s", self.task_id, timeout)
        logging.info("Command: %s" % cmd)
        timeout = int(timeout)

        if not self.history_storage.task_valid_to_run(self.task_id):
            raise FileExistsError('%s task is in progress' % self.task_id)

        self.task_run = TaskRun(self.task_id)
        self.fails_before_notify = int(fails_before_notify)
        self.history_storage.prepare_task_run(self.task_run)
        self.history_storage.update_state_for_task_run(self.task_run)

        self.proc = AASubprocess(cmd, timeout, self.chunk_stdout, self.chunk_stderr,
                                 self.done_callback, kill_on_timeout=True)
        self.task_run.state['pid'] = self.proc.pid
        self.history_storage.update_state_for_task_run(self.task_run)

    def chunk_stdout(self, data):
        self.task_run.stdout += data.decode()
        self.history_storage.update_stdout_for_task_run_id(self.task_run)

    def chunk_stderr(self, data):
        self.task_run.stderr += data.decode()
        self.history_storage.update_stderr_for_task_run_id(self.task_run)

    def done_callback(self, return_code, expired):
        self.task_run.state['exit_code'] = return_code
        self.task_run.finalize()
        self.history_storage.update_state_for_task_run(self.task_run)

        if self.datadog_config['enabled']:
            task_time = (self.task_run.finished_at - self.task_run.started_at).seconds
            self.datadog_config['statsd_client'].gauge('ereb.%s' % self.task_id, task_time, tags=[self.datadog_config['tag']])

        if return_code != 0:
            last_fails = self.history_storage.last_fails(self.task_id)
            if self.fails_before_notify == 0 or (last_fails % self.fails_before_notify == 0):
                self.on_error_callback(self.task_id, return_code)
                self.notifier.send_failed_task_run(self.task_run, last_fails)



import time
import json
import glob
import re
import uuid
from crontab import CronTab
from tornado.ioloop import IOLoop, PeriodicCallback
from tornado import gen
import logging
from os.path import isfile
from datadog import statsd

from ereb.task_runner import TaskRunner


class TasksScheduler():
    SHELL_SCRIPT_RE = r'(\S+\.(sh|rb|py))'

    def __init__(self, tasks_dir, history_storage, notifier, datadog_config):
        self.tasks_dir = tasks_dir
        self.history_storage = history_storage
        self.notifier = notifier
        self.datadog_config = datadog_config
        self.tasks_list = []
        self.is_task_loop_running = False
        self.planned_task_run_uuids = []
        self.try_after_fail_tasks = {}
        self.try_after_fail_tries_count = 2
        self.try_after_fail_interval = 60  # 6
        self.task_queue_by_timestamp = {}
        self.update_config()

    def update_config(self):
        new_config = self.get_tasks_config()
        old_tasks_by_id = {}
        new_tasks_by_id = {}
        for old_task in self.tasks_list:
            old_tasks_by_id[old_task['name']] = {
                'cron_schedule': old_task.get('cron_schedule'),
                'cmd': old_task.get('cmd'),
                'try_more_on_error': old_task.get('try_more_on_error')
            }
        for new_task in new_config:
            new_tasks_by_id[new_task['name']] = {
                'cron_schedule': new_task.get('cron_schedule'),
                'cmd': new_task.get('cmd'),
                'try_more_on_error': new_task.get('try_more_on_error')
            }

        if new_tasks_by_id != old_tasks_by_id:
            self.tasks_list = new_config
            return True

    def get_task_config(self, name):
        for config in self.tasks_list:
            if config['name'] == name:
                return config
        return None

    def start(self):
        self.config_checking_loop = PeriodicCallback(self.check_config, 1 * 1000)
        self.config_checking_loop.start()
        self.start_task_loop()

    def start_task_loop(self):
        self.is_task_loop_running = True
        IOLoop.instance().add_callback(self.schedule_next_tasks)

    def stop_task_loop(self):
        self.is_task_loop_running = False
        self.planned_task_run_uuids = []

    def run_task_by_name_and_cmd(self, name, cmd, timeout, fails_before_notify=0):
        logging.info('Manual run | Running %s task' % name)
        try:
            task_runner = TaskRunner(name, self.history_storage, self.notifier,
                                     self.on_task_fail_callback, self.datadog_config)
            task_runner.run_task(cmd, timeout, fails_before_notify)
        except Exception as e:
            logging.error('Manual task run error. %s' % e)

    def on_task_fail_callback(self, task_id, return_code):
        def add_failed_task_to_queue(task_id):
            next_run = time.time() + self.try_after_fail_interval
            if next_run in self.task_queue_by_timestamp:
                self.task_queue_by_timestamp[next_run].append(task_id)
            else:
                self.task_queue_by_timestamp[next_run] = [task_id]

        task = self.get_task_by_id(task_id, False)
        if task and task.get('try_more_on_error', False):
            if task_id in self.try_after_fail_tasks:
                self.try_after_fail_tasks[task_id] -= 1
                if self.try_after_fail_tasks[task_id] == 0:
                    # no more tries
                    self.try_after_fail_tasks.pop(task_id)
                    self.reschedule_tasks()
                else:
                    add_failed_task_to_queue(task_id)
                    self.reschedule_tasks()
            else:
                self.try_after_fail_tasks[task_id] = self.try_after_fail_tries_count

                add_failed_task_to_queue(task_id)
                self.reschedule_tasks()

    @gen.engine
    def check_config(self):
        if self.update_config():
            logging.info("Config changed")
            self.reschedule_tasks()

    @gen.engine
    def reschedule_tasks(self):
        logging.info("Recheduling tasks!")
        self.planned_task_run_uuids = []
        IOLoop.instance().add_callback(self.schedule_next_tasks)

    def get_tasks_config(self):
        # async?
        regexp = re.compile('.+\/(.+).json', re.IGNORECASE)
        config = []
        for f in glob.glob(self.tasks_dir + '/*.json'):
            try:
                task_name = regexp.search(f).group(1)
                with open(f) as config_file:
                    c = json.load(config_file)
                if self.validate_config(c):
                    c['name'] = task_name
                    config.append(c)
                else:
                    logging.info("Something bad with %s config file" % f)
            except Exception:
                logging.info("Error loading %s config file" % f)

        return config

    def validate_config(self, config):
        try:
            if 'cron_schedule' in config:
                next_time = CronTab(config['cron_schedule']).next(default_utc=True)
            result = isinstance(config, dict)
        except:
            logging.info("BadConfigException: %s" % config)
            result = False

        return result

    def get_status(self):
        result = {}
        if self.is_task_loop_running:
            result['state'] = 'running'
        else:
            result['state'] = 'stopped'

        result['next_run'], result['next_tasks'] = self.get_next_tasks()
        result['planned_task_run_uuids'] = self.planned_task_run_uuids

        return result

    @gen.engine
    def schedule_next_tasks(self):
        if self.is_task_loop_running:
            logging.info("Tasks loop started")
            self.notifier.websocket_send_status(self.get_status())
            next_run, next_tasks = self.get_next_tasks()
            if len(next_tasks) > 0:
                logging.info('Next run in %s seconds' % str(next_run))
                task_run_uuid = str(uuid.uuid4())
                self.planned_task_run_uuids.append(task_run_uuid)  # allows to cancel already scheduled tasks
                logging.info('Planned task %s' % task_run_uuid)
                # wait until next task needs to be run, and while waiting -
                # do other stuff in IOLoop (i.e. process web interface requests)
                yield gen.Task(IOLoop.instance().add_timeout, time.time() + next_run)
                if task_run_uuid in self.planned_task_run_uuids:
                    logging.info('Now running %s tasks' % len(next_tasks))
                    for task in next_tasks:
                        try:
                            logging.info('Running %s task with timeout %s', task['name'], task.get('timeout', -1))
                            task_runner = TaskRunner(task['name'], self.history_storage,
                                                     self.notifier, self.on_task_fail_callback, self.datadog_config)
                            task_runner.run_task(task['cmd'], task.get('timeout', -1),
                                                 task.get('fails_before_notify', 0))
                        except Exception as e:
                            logging.exception('Scheduled task run error %s' % task['name'])
                    self.planned_task_run_uuids.remove(task_run_uuid)
                    logging.info('Run and removed task run %s' % task_run_uuid)
                    IOLoop.instance().add_callback(self.schedule_next_tasks)
                else:
                    logging.info('Task %s run %s was cancelled for ' % (task['name'], task_run_uuid))
            self.clean_task_queue_by_timestamp()
        else:
            logging.info("TaskRunner stopped")

    def get_next_tasks(self):
        tasks_by_schedule = {}
        now = time.time()

        for task in self.tasks_list:
            if task.get('enabled', False):
                next = CronTab(task['cron_schedule']).next(now, default_utc=True)
                if next in tasks_by_schedule:
                    tasks_by_schedule[next].append(task)
                else:
                    tasks_by_schedule[next] = [task]

        for next_timestamp, tasks in self.task_queue_by_timestamp.items():
            next = next_timestamp - time.time()
            for task_id in tasks:
                task = self.get_task_by_id(task_id, False)
                if next in tasks_by_schedule:
                    tasks_by_schedule[next].append(task)
                else:
                    tasks_by_schedule[next] = [task]

        if len(tasks_by_schedule) > 0:
            return sorted(tasks_by_schedule.items(), key=lambda x: x[0])[0]
        else:
            return [0, []]

    def get_task_by_id(self, task_id, with_extra_info=False):
        for task in self.tasks_list:
            if task['name'] == task_id:
                task['shell_scripts'] = []
                if with_extra_info:
                    task['shell_scripts'] = self.try_to_parse_task_shell_script(task['cmd'])

                return task

        return None

    def clean_task_queue_by_timestamp(self):
        for next_timestamp in list(self.task_queue_by_timestamp.keys()):
            if time.time() > next_timestamp:
                del self.task_queue_by_timestamp[next_timestamp]

    def try_to_parse_task_shell_script(self, cmd):
        scripts = list(map(lambda x: x[0], re.findall(self.SHELL_SCRIPT_RE, cmd)))

        def read_file(shell_script):
            with open(shell_script, 'r', encoding='utf8') as content:
                return {'filename': shell_script, 'content': content.read()}
        return_data = []
        for script in scripts:
            if isfile(script):
                return_data.append(read_file(script))
        return return_data



from setuptools import setup, find_packages

setup(
    name='ereb',
    description='Tornado based cron with web interface, JSON API and history',
    version='1.1.1',
    url='https://github.com/KosyanMedia/ereb',
    author='Alex Shaikhaleev',
    author_email='nimdraug.sael@gmail.com',
    keywords='tornado web cron',
    packages=['ereb'],
    install_requires=['tornado==4.3', 'crontab==0.22', 'psutil', 'requests', 'datadog'],
    include_package_data=True,
    license='MIT',
    entry_points="""
      [console_scripts]
      ereb = ereb.erebd:main
      """
)



import subprocess
import urllib.parse
import logging
import requests


class Notifier():
    def __init__(self, notifier_config, notify_to='logger', websocket_clients=[], host='hostname', port=8888):
        self.notify_to = notify_to
        if host == 'hostname':
            self.hostname = self.get_hostname()
        else:
            self.hostname = host
        self.port = port
        self.websocket_clients = websocket_clients
        try:
            if notify_to in notifier_config:
                self.cmd = notifier_config[notify_to]
            else:
                self.cmd = None
        except:
            logging.exception("Notifier error:")
            self.cmd = None

    def error(self, link, message, task_run, fails):
        if self.cmd:
            if self.notify_to.startswith('slack_api'):
                info = task_run.log_info(fails, lines_count=self.cmd.get('include_strings', 2))
                slack_payload = {
                    'attachments': [{
                        'pretext': 'Failed task',
                        'title': info['task_id'],
                        'title_link': link,
                        'text': self.cmd.get('text', 'Failed task {task_id}').format(**info),
                        'fallback': message,
                        'short': True,
                        'color': 'danger',
                        'mrkdwn_in': ['text', 'pretext']
                    }]
                }
                channel = self.cmd.get('channel', False)
                if channel:
                    slack_payload['channel'] = channel
                response = requests.post(self.cmd['webhook_url'], json=slack_payload)
                if response.status_code != 200:
                    logging.error('Notifications Slack API error: %s' % response.text)
            else:
                if "http" in self.cmd:
                    link = urllib.parse.quote(link)
                    message = urllib.parse.quote(message)

                notifier_cmd = self.cmd.replace('%s', "{0}\n{1}".format(link, message))
                subprocess.Popen(notifier_cmd, shell=True, stdout=subprocess.PIPE)
        else:
            logging.warning("Notifications are turned off")

    def send_failed_task_run(self, task_run, fails):
        link = "http://{0}:{1}/#/tasks/{2}/runs/{3}".format(self.hostname, self.port, task_run.task_id, task_run.id)
        if fails > 1:
            message = "Task {0} failed {1} times".format(task_run.task_id, str(fails))
        else:
            message = "Task {0} failed".format(task_run.task_id)
        self.error(link, message, task_run, fails)

    def get_hostname(self):
        return subprocess.Popen("hostname", shell=True, stdout=subprocess.PIPE).stdout.read().decode().replace('\n', '')

    def websocket_send_status(self, message):
        if len(self.websocket_clients) > 0:
            logging.info('websocket_send_status')
            for client in self.websocket_clients:
                client.write_message(message)
